@@ -4,10 +4,10 @@
  *
  * PostgreSQL object comments utility code.
  *
- * Copyright (c) 1996-2001, PostgreSQL Global Development Group
+ * Copyright (c) 1996-2003, PostgreSQL Global Development Group
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql-server/src/backend/commands/comment.c,v 1.61.2.1 2003/07/17 20:14:09 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/commands/comment.c,v 1.71 2003/09/25 06:57:58 petere Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -71,45 +71,45 @@ CommentObject(CommentStmt *stmt)
 {
 	switch (stmt->objtype)
 	{
-		case COMMENT_ON_INDEX:
-		case COMMENT_ON_SEQUENCE:
-		case COMMENT_ON_TABLE:
-		case COMMENT_ON_VIEW:
+		case OBJECT_INDEX:
+		case OBJECT_SEQUENCE:
+		case OBJECT_TABLE:
+		case OBJECT_VIEW:
 			CommentRelation(stmt->objtype, stmt->objname, stmt->comment);
 			break;
-		case COMMENT_ON_COLUMN:
+		case OBJECT_COLUMN:
 			CommentAttribute(stmt->objname, stmt->comment);
 			break;
-		case COMMENT_ON_DATABASE:
+		case OBJECT_DATABASE:
 			CommentDatabase(stmt->objname, stmt->comment);
 			break;
-		case COMMENT_ON_RULE:
+		case OBJECT_RULE:
 			CommentRule(stmt->objname, stmt->comment);
 			break;
-		case COMMENT_ON_TYPE:
+		case OBJECT_TYPE:
 			CommentType(stmt->objname, stmt->comment);
 			break;
-		case COMMENT_ON_AGGREGATE:
+		case OBJECT_AGGREGATE:
 			CommentAggregate(stmt->objname, stmt->objargs, stmt->comment);
 			break;
-		case COMMENT_ON_FUNCTION:
+		case OBJECT_FUNCTION:
 			CommentProc(stmt->objname, stmt->objargs, stmt->comment);
 			break;
-		case COMMENT_ON_OPERATOR:
+		case OBJECT_OPERATOR:
 			CommentOperator(stmt->objname, stmt->objargs, stmt->comment);
 			break;
-		case COMMENT_ON_TRIGGER:
+		case OBJECT_TRIGGER:
 			CommentTrigger(stmt->objname, stmt->comment);
 			break;
-		case COMMENT_ON_SCHEMA:
+		case OBJECT_SCHEMA:
 			CommentNamespace(stmt->objname, stmt->comment);
 			break;
-		case COMMENT_ON_CONSTRAINT:
+		case OBJECT_CONSTRAINT:
 			CommentConstraint(stmt->objname, stmt->comment);
 			break;
 		default:
-			elog(ERROR, "An attempt was made to comment on a unknown type: %d",
-				 stmt->objtype);
+			elog(ERROR, "unrecognized object type: %d",
+				 (int) stmt->objtype);
 	}
 }
 
@@ -295,31 +295,40 @@ CommentRelation(int objtype, List *relname, char *comment)
 
 	/* Check object security */
 	if (!pg_class_ownercheck(RelationGetRelid(relation), GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, RelationGetRelationName(relation));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
+					   RelationGetRelationName(relation));
 
 	/* Next, verify that the relation type matches the intent */
 
 	switch (objtype)
 	{
-		case COMMENT_ON_INDEX:
+		case OBJECT_INDEX:
 			if (relation->rd_rel->relkind != RELKIND_INDEX)
-				elog(ERROR, "relation \"%s\" is not an index",
-					 RelationGetRelationName(relation));
+				ereport(ERROR,
+						(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+						 errmsg("\"%s\" is not an index",
+								RelationGetRelationName(relation))));
 			break;
-		case COMMENT_ON_SEQUENCE:
+		case OBJECT_SEQUENCE:
 			if (relation->rd_rel->relkind != RELKIND_SEQUENCE)
-				elog(ERROR, "relation \"%s\" is not a sequence",
-					 RelationGetRelationName(relation));
+				ereport(ERROR,
+						(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+						 errmsg("\"%s\" is not a sequence",
+								RelationGetRelationName(relation))));
 			break;
-		case COMMENT_ON_TABLE:
+		case OBJECT_TABLE:
 			if (relation->rd_rel->relkind != RELKIND_RELATION)
-				elog(ERROR, "relation \"%s\" is not a table",
-					 RelationGetRelationName(relation));
+				ereport(ERROR,
+						(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+						 errmsg("\"%s\" is not a table",
+								RelationGetRelationName(relation))));
 			break;
-		case COMMENT_ON_VIEW:
+		case OBJECT_VIEW:
 			if (relation->rd_rel->relkind != RELKIND_VIEW)
-				elog(ERROR, "relation \"%s\" is not a view",
-					 RelationGetRelationName(relation));
+				ereport(ERROR,
+						(errcode(ERRCODE_WRONG_OBJECT_TYPE),
+						 errmsg("\"%s\" is not a view",
+								RelationGetRelationName(relation))));
 			break;
 	}
 
@@ -338,8 +347,8 @@ CommentRelation(int objtype, List *relname, char *comment)
  * such as a table's column. The routine will check security
  * restrictions and then attempt to look up the specified
  * attribute. If successful, a comment is added/dropped, else an
- * elog() exception is thrown.	The parameters are the relation
- * and attribute names, and the comments
+ * ereport() exception is thrown.	The parameters are the relation
+ * and attribute names, and the comment
  */
 static void
 CommentAttribute(List *qualname, char *comment)
@@ -353,10 +362,10 @@ CommentAttribute(List *qualname, char *comment)
 
 	/* Separate relname and attr name */
 	nnames = length(qualname);
-	if (nnames < 2)
-		elog(ERROR, "CommentAttribute: must specify relation.attribute");
+	if (nnames < 2)				/* parser messed up */
+		elog(ERROR, "must specify relation and attribute");
 	relname = ltruncate(nnames - 1, listCopy(qualname));
-	attrname = strVal(nth(nnames - 1, qualname));
+	attrname = strVal(llast(qualname));
 
 	/* Open the containing relation to ensure it won't go away meanwhile */
 	rel = makeRangeVarFromNameList(relname);
@@ -365,14 +374,17 @@ CommentAttribute(List *qualname, char *comment)
 	/* Check object security */
 
 	if (!pg_class_ownercheck(RelationGetRelid(relation), GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, RelationGetRelationName(relation));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
+					   RelationGetRelationName(relation));
 
 	/* Now, fetch the attribute number from the system cache */
 
 	attnum = get_attnum(RelationGetRelid(relation), attrname);
 	if (attnum == InvalidAttrNumber)
-		elog(ERROR, "Relation \"%s\" has no column \"%s\"",
-			 RelationGetRelationName(relation), attrname);
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_COLUMN),
+			 errmsg("column \"%s\" of relation \"%s\" does not exist",
+					attrname, RelationGetRelationName(relation))));
 
 	/* Create the comment using the relation's oid */
 
@@ -389,7 +401,7 @@ CommentAttribute(List *qualname, char *comment)
  *
  * This routine is used to add/drop any user-comments a user might
  * have regarding the specified database. The routine will check
- * security for owner permissions, and, if succesful, will then
+ * security for owner permissions, and, if successful, will then
  * attempt to find the oid of the database specified. Once found,
  * a comment is added/dropped using the CreateComments() routine.
  */
@@ -400,41 +412,48 @@ CommentDatabase(List *qualname, char *comment)
 	Oid			oid;
 
 	if (length(qualname) != 1)
-		elog(ERROR, "CommentDatabase: database name may not be qualified");
+		ereport(ERROR,
+				(errcode(ERRCODE_SYNTAX_ERROR),
+				 errmsg("database name may not be qualified")));
 	database = strVal(lfirst(qualname));
 
 	/*
-	 * We cannot currently support cross-database comments (since other DBs
-	 * cannot see pg_description of this database).  So, we reject attempts
-	 * to comment on a database other than the current one.  Someday this
-	 * might be improved, but it would take a redesigned infrastructure.
+	 * We cannot currently support cross-database comments (since other
+	 * DBs cannot see pg_description of this database).  So, we reject
+	 * attempts to comment on a database other than the current one.
+	 * Someday this might be improved, but it would take a redesigned
+	 * infrastructure.
 	 *
 	 * When loading a dump, we may see a COMMENT ON DATABASE for the old name
-	 * of the database.  Erroring out would prevent pg_restore from completing
-	 * (which is really pg_restore's fault, but for now we will work around
-	 * the problem here).  Consensus is that the best fix is to treat wrong
-	 * database name as a WARNING not an ERROR.
+	 * of the database.  Erroring out would prevent pg_restore from
+	 * completing (which is really pg_restore's fault, but for now we will
+	 * work around the problem here).  Consensus is that the best fix is
+	 * to treat wrong database name as a WARNING not an ERROR.
 	 */
 
 	/* First get the database OID */
 	oid = get_database_oid(database);
 	if (!OidIsValid(oid))
 	{
-		elog(WARNING, "database \"%s\" does not exist", database);
+		ereport(WARNING,
+				(errcode(ERRCODE_UNDEFINED_DATABASE),
+				 errmsg("database \"%s\" does not exist", database)));
 		return;
 	}
 
 	/* Only allow comments on the current database */
 	if (oid != MyDatabaseId)
 	{
-		elog(WARNING, "Database comments may only be applied to the current database");
+		ereport(WARNING,
+				(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+				 errmsg("database comments may only be applied to the current database")));
 		return;
 	}
 
-	/* Allow if the user matches the database dba or is a superuser */
-	if (!(superuser() || is_dbadmin(oid)))
-		elog(ERROR, "you are not permitted to comment on database \"%s\"",
-			 database);
+	/* Check object security */
+	if (!pg_database_ownercheck(oid, GetUserId()))
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_DATABASE,
+					   database);
 
 	/* Create the comment with the pg_database oid */
 	CreateComments(oid, RelOid_pg_database, 0, comment);
@@ -445,7 +464,7 @@ CommentDatabase(List *qualname, char *comment)
  *
  * This routine is used to add/drop any user-comments a user might
  * have regarding the specified namespace. The routine will check
- * security for owner permissions, and, if succesful, will then
+ * security for owner permissions, and, if successful, will then
  * attempt to find the oid of the namespace specified. Once found,
  * a comment is added/dropped using the CreateComments() routine.
  */
@@ -457,19 +476,23 @@ CommentNamespace(List *qualname, char *comment)
 	char	   *namespace;
 
 	if (length(qualname) != 1)
-		elog(ERROR, "CommentSchema: schema name may not be qualified");
+		ereport(ERROR,
+				(errcode(ERRCODE_SYNTAX_ERROR),
+				 errmsg("schema name may not be qualified")));
 	namespace = strVal(lfirst(qualname));
 
 	oid = GetSysCacheOid(NAMESPACENAME,
 						 CStringGetDatum(namespace),
 						 0, 0, 0);
 	if (!OidIsValid(oid))
-		elog(ERROR, "CommentSchema: Schema \"%s\" could not be found",
-			 namespace);
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_SCHEMA),
+				 errmsg("schema \"%s\" does not exist", namespace)));
 
 	/* Check object security */
 	if (!pg_namespace_ownercheck(oid, GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, namespace);
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_NAMESPACE,
+					   namespace);
 
 	/* pg_namespace doesn't have a hard-coded OID, so must look it up */
 	classoid = get_system_catalog_relid(NamespaceRelationName);
@@ -536,15 +559,18 @@ CommentRule(List *qualname, char *comment)
 		}
 		else
 		{
-			elog(ERROR, "rule \"%s\" does not exist", rulename);
+			ereport(ERROR,
+					(errcode(ERRCODE_UNDEFINED_OBJECT),
+					 errmsg("rule \"%s\" does not exist", rulename)));
 			reloid = ruleoid = 0;		/* keep compiler quiet */
 		}
 
 		if (HeapTupleIsValid(tuple = heap_getnext(scanDesc,
 												  ForwardScanDirection)))
-			elog(ERROR, "There are multiple rules \"%s\""
-			 "\n\tPlease specify a relation name as well as a rule name",
-				 rulename);
+			ereport(ERROR,
+					(errcode(ERRCODE_DUPLICATE_OBJECT),
+					 errmsg("there are multiple rules named \"%s\"", rulename),
+			errhint("Specify a relation name as well as a rule name.")));
 
 		heap_endscan(scanDesc);
 		heap_close(RewriteRelation, AccessShareLock);
@@ -557,7 +583,7 @@ CommentRule(List *qualname, char *comment)
 		/* New-style: rule and relname both provided */
 		Assert(nnames >= 2);
 		relname = ltruncate(nnames - 1, listCopy(qualname));
-		rulename = strVal(nth(nnames - 1, qualname));
+		rulename = strVal(llast(qualname));
 
 		/* Open the owning relation to ensure it won't go away meanwhile */
 		rel = makeRangeVarFromNameList(relname);
@@ -570,7 +596,7 @@ CommentRule(List *qualname, char *comment)
 							   PointerGetDatum(rulename),
 							   0, 0);
 		if (!HeapTupleIsValid(tuple))
-			elog(ERROR, "rule \"%s\" does not exist", rulename);
+			elog(ERROR, "cache lookup failed for rule \"%s\"", rulename);
 		Assert(reloid == ((Form_pg_rewrite) GETSTRUCT(tuple))->ev_class);
 		ruleoid = HeapTupleGetOid(tuple);
 		ReleaseSysCache(tuple);
@@ -579,7 +605,8 @@ CommentRule(List *qualname, char *comment)
 	/* Check object security */
 	aclcheck = pg_class_aclcheck(reloid, GetUserId(), ACL_RULE);
 	if (aclcheck != ACLCHECK_OK)
-		aclcheck_error(aclcheck, rulename);
+		aclcheck_error(aclcheck, ACL_KIND_CLASS,
+					   get_rel_name(reloid));
 
 	/* pg_rewrite doesn't have a hard-coded OID, so must look it up */
 	classoid = get_system_catalog_relid(RewriteRelationName);
@@ -597,7 +624,7 @@ CommentRule(List *qualname, char *comment)
  * have regarding a TYPE. The type is specified by name
  * and, if found, and the user has appropriate permissions, a
  * comment will be added/dropped using the CreateComments() routine.
- * The type's name and the comments are the paramters to this routine.
+ * The type's name and the comments are the parameters to this routine.
  */
 static void
 CommentType(List *typename, char *comment)
@@ -617,7 +644,8 @@ CommentType(List *typename, char *comment)
 	/* Check object security */
 
 	if (!pg_type_ownercheck(oid, GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, TypeNameToString(tname));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_TYPE,
+					   TypeNameToString(tname));
 
 	/* Call CreateComments() to create/drop the comments */
 
@@ -647,12 +675,13 @@ CommentAggregate(List *aggregate, List *arguments, char *comment)
 
 	/* Now, attempt to find the actual tuple in pg_proc */
 
-	oid = find_aggregate_func("CommentAggregate", aggregate, baseoid);
+	oid = find_aggregate_func(aggregate, baseoid, false);
 
 	/* Next, validate the user's attempt to comment */
 
 	if (!pg_proc_ownercheck(oid, GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, NameListToString(aggregate));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_PROC,
+					   NameListToString(aggregate));
 
 	/* Call CreateComments() to create/drop the comments */
 
@@ -675,13 +704,13 @@ CommentProc(List *function, List *arguments, char *comment)
 
 	/* Look up the procedure */
 
-	oid = LookupFuncNameTypeNames(function, arguments,
-								  "CommentProc");
+	oid = LookupFuncNameTypeNames(function, arguments, false);
 
 	/* Now, validate the user's ability to comment on this function */
 
 	if (!pg_proc_ownercheck(oid, GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, NameListToString(function));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_PROC,
+					   NameListToString(function));
 
 	/* Call CreateComments() to create/drop the comments */
 
@@ -707,12 +736,12 @@ CommentOperator(List *opername, List *arguments, char *comment)
 	Oid			classoid;
 
 	/* Look up the operator */
-	oid = LookupOperNameTypeNames(opername, typenode1, typenode2,
-								  "CommentOperator");
+	oid = LookupOperNameTypeNames(opername, typenode1, typenode2, false);
 
 	/* Valid user's ability to comment on this operator */
 	if (!pg_oper_ownercheck(oid, GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, NameListToString(opername));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_OPER,
+					   NameListToString(opername));
 
 	/* pg_operator doesn't have a hard-coded OID, so must look it up */
 	classoid = get_system_catalog_relid(OperatorRelationName);
@@ -746,10 +775,10 @@ CommentTrigger(List *qualname, char *comment)
 
 	/* Separate relname and trig name */
 	nnames = length(qualname);
-	if (nnames < 2)
-		elog(ERROR, "CommentTrigger: must specify relation and trigger");
+	if (nnames < 2)				/* parser messed up */
+		elog(ERROR, "must specify relation and trigger");
 	relname = ltruncate(nnames - 1, listCopy(qualname));
-	trigname = strVal(nth(nnames - 1, qualname));
+	trigname = strVal(llast(qualname));
 
 	/* Open the owning relation to ensure it won't go away meanwhile */
 	rel = makeRangeVarFromNameList(relname);
@@ -758,7 +787,8 @@ CommentTrigger(List *qualname, char *comment)
 	/* Check object security */
 
 	if (!pg_class_ownercheck(RelationGetRelid(relation), GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, RelationGetRelationName(relation));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
+					   RelationGetRelationName(relation));
 
 	/*
 	 * Fetch the trigger tuple from pg_trigger.  There can be only one
@@ -780,8 +810,10 @@ CommentTrigger(List *qualname, char *comment)
 	/* If no trigger exists for the relation specified, notify user */
 
 	if (!HeapTupleIsValid(triggertuple))
-		elog(ERROR, "trigger \"%s\" for relation \"%s\" does not exist",
-			 trigname, RelationGetRelationName(relation));
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+			  errmsg("trigger \"%s\" for table \"%s\" does not exist",
+					 trigname, RelationGetRelationName(relation))));
 
 	oid = HeapTupleGetOid(triggertuple);
 
@@ -821,10 +853,10 @@ CommentConstraint(List *qualname, char *comment)
 
 	/* Separate relname and constraint name */
 	nnames = length(qualname);
-	if (nnames < 2)
-		elog(ERROR, "CommentConstraint: must specify relation and constraint");
+	if (nnames < 2)				/* parser messed up */
+		elog(ERROR, "must specify relation and constraint");
 	relName = ltruncate(nnames - 1, listCopy(qualname));
-	conName = strVal(nth(nnames - 1, qualname));
+	conName = strVal(llast(qualname));
 
 	/* Open the owning relation to ensure it won't go away meanwhile */
 	rel = makeRangeVarFromNameList(relName);
@@ -833,7 +865,8 @@ CommentConstraint(List *qualname, char *comment)
 	/* Check object security */
 
 	if (!pg_class_ownercheck(RelationGetRelid(relation), GetUserId()))
-		aclcheck_error(ACLCHECK_NOT_OWNER, RelationGetRelationName(relation));
+		aclcheck_error(ACLCHECK_NOT_OWNER, ACL_KIND_CLASS,
+					   RelationGetRelationName(relation));
 
 	/*
 	 * Fetch the constraint tuple from pg_constraint.  There may be more
@@ -856,8 +889,10 @@ CommentConstraint(List *qualname, char *comment)
 		if (strcmp(NameStr(con->conname), conName) == 0)
 		{
 			if (OidIsValid(conOid))
-				elog(ERROR, "Relation \"%s\" has multiple constraints named \"%s\"",
-					 RelationGetRelationName(relation), conName);
+				ereport(ERROR,
+						(errcode(ERRCODE_DUPLICATE_OBJECT),
+						 errmsg("table \"%s\" has multiple constraints named \"%s\"",
+						   RelationGetRelationName(relation), conName)));
 			conOid = HeapTupleGetOid(tuple);
 		}
 	}
@@ -866,8 +901,10 @@ CommentConstraint(List *qualname, char *comment)
 
 	/* If no constraint exists for the relation specified, notify user */
 	if (!OidIsValid(conOid))
-		elog(ERROR, "constraint \"%s\" for relation \"%s\" does not exist",
-			 conName, RelationGetRelationName(relation));
+		ereport(ERROR,
+				(errcode(ERRCODE_UNDEFINED_OBJECT),
+		   errmsg("constraint \"%s\" for table \"%s\" does not exist",
+				  conName, RelationGetRelationName(relation))));
 
 	/* Create the comment with the pg_constraint oid */
 	CreateComments(conOid, RelationGetRelid(pg_constraint), 0, comment);
