@@ -3,12 +3,12 @@
  * date.c
  *	  implements DATE and TIME data types specified in SQL-92 standard
  *
- * Portions Copyright (c) 1996-2002, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2003, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994-5, Regents of the University of California
  *
  *
  * IDENTIFICATION
- *	  $Header: /cvsroot/pgsql/src/backend/utils/adt/date.c,v 1.73.2.8 2005/04/23 22:54:03 tgl Exp $
+ *	  $Header: /cvsroot/pgsql-server/src/backend/utils/adt/date.c,v 1.93.2.1 2004/06/13 17:17:48 tgl Exp $
  *
  *-------------------------------------------------------------------------
  */
@@ -21,6 +21,7 @@
 #include <float.h>
 
 #include "access/hash.h"
+#include "libpq/pqformat.h"
 #include "miscadmin.h"
 #include "utils/builtins.h"
 #include "utils/date.h"
@@ -61,16 +62,19 @@ date_in(PG_FUNCTION_ARGS)
 	int			tzp;
 	int			dtype;
 	int			nf;
+	int			dterr;
 	char	   *field[MAXDATEFIELDS];
 	int			ftype[MAXDATEFIELDS];
 	char		lowstr[MAXDATELEN + 1];
 
 	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad date external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-	 || (DecodeDateTime(field, ftype, nf, &dtype, tm, &fsec, &tzp) != 0))
-		elog(ERROR, "Bad date external representation '%s'", str);
+		dterr = DTERR_BAD_FORMAT;
+	else
+		dterr = ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeDateTime(field, ftype, nf, &dtype, tm, &fsec, &tzp);
+	if (dterr != 0)
+		DateTimeParseError(dterr, str, "date");
 
 	switch (dtype)
 	{
@@ -78,8 +82,10 @@ date_in(PG_FUNCTION_ARGS)
 			break;
 
 		case DTK_CURRENT:
-			elog(ERROR, "Date CURRENT no longer supported"
-				 "\n\tdate_in() internal coding error");
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+					 errmsg("date/time value \"current\" is no longer supported")));
+
 			GetCurrentDateTime(tm);
 			break;
 
@@ -88,10 +94,11 @@ date_in(PG_FUNCTION_ARGS)
 			break;
 
 		default:
-			elog(ERROR, "Unrecognized date external representation '%s'", str);
+			DateTimeParseError(DTERR_BAD_FORMAT, str, "date");
+			break;
 	}
 
-	date = (date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1));
+	date = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 
 	PG_RETURN_DATEADT(date);
 }
@@ -108,7 +115,7 @@ date_out(PG_FUNCTION_ARGS)
 			   *tm = &tt;
 	char		buf[MAXDATELEN + 1];
 
-	j2date((date + date2j(2000, 1, 1)),
+	j2date(date + POSTGRES_EPOCH_JDATE,
 		   &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
 
 	EncodeDateOnly(tm, DateStyle, buf);
@@ -117,6 +124,32 @@ date_out(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		date_recv			- converts external binary format to date
+ */
+Datum
+date_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+
+	PG_RETURN_DATEADT((DateADT) pq_getmsgint(buf, sizeof(DateADT)));
+}
+
+/*
+ *		date_send			- converts date to binary format
+ */
+Datum
+date_send(PG_FUNCTION_ARGS)
+{
+	DateADT		date = PG_GETARG_DATEADT(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+	pq_sendint(&buf, date, sizeof(date));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
+
 Datum
 date_eq(PG_FUNCTION_ARGS)
 {
@@ -256,9 +289,10 @@ date_pl_interval(PG_FUNCTION_ARGS)
 
 	if (span->month != 0)
 	{
-		j2date((dateVal + date2j(2000, 1, 1)), &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
+		j2date(dateVal + POSTGRES_EPOCH_JDATE,
+			   &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
 		tm->tm_mon += span->month;
-		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 	}
 	if (span->time != 0)
 		dateVal += (span->time / 86400e0);
@@ -279,9 +313,10 @@ date_mi_interval(PG_FUNCTION_ARGS)
 
 	if (span->month != 0)
 	{
-		j2date((dateVal + date2j(2000, 1, 1)), &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
+		j2date(dateVal + POSTGRES_EPOCH_JDATE,
+			   &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
 		tm->tm_mon -= span->month;
-		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+		dateVal = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 	}
 	if (span->time != 0)
 		dateVal -= (span->time / 86400e0);
@@ -327,9 +362,11 @@ timestamp_date(PG_FUNCTION_ARGS)
 		PG_RETURN_NULL();
 
 	if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL) != 0)
-		elog(ERROR, "Unable to convert timestamp to date");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
-	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 
 	PG_RETURN_DATEADT(result);
 }
@@ -346,7 +383,7 @@ date_timestamptz(PG_FUNCTION_ARGS)
 	struct tm	tt,
 			   *tm = &tt;
 
-	j2date((dateVal + date2j(2000, 1, 1)),
+	j2date(dateVal + POSTGRES_EPOCH_JDATE,
 		   &(tm->tm_year), &(tm->tm_mon), &(tm->tm_mday));
 
 	if (IS_VALID_UTIME(tm->tm_year, tm->tm_mon, tm->tm_mday))
@@ -397,9 +434,11 @@ timestamptz_date(PG_FUNCTION_ARGS)
 		PG_RETURN_NULL();
 
 	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
-		elog(ERROR, "Unable to convert timestamp to date");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
-	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+	result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 
 	PG_RETURN_DATEADT(result);
 }
@@ -422,7 +461,9 @@ abstime_date(PG_FUNCTION_ARGS)
 		case INVALID_ABSTIME:
 		case NOSTART_ABSTIME:
 		case NOEND_ABSTIME:
-			elog(ERROR, "Unable to convert reserved abstime value to date");
+			ereport(ERROR,
+					(errcode(ERRCODE_FEATURE_NOT_SUPPORTED),
+			   errmsg("cannot convert reserved abstime value to date")));
 
 			/*
 			 * pretend to drop through to make compiler think that result
@@ -431,7 +472,7 @@ abstime_date(PG_FUNCTION_ARGS)
 
 		default:
 			abstime2tm(abstime, &tz, tm, NULL);
-			result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - date2j(2000, 1, 1);
+			result = date2j(tm->tm_year, tm->tm_mon, tm->tm_mday) - POSTGRES_EPOCH_JDATE;
 			break;
 	}
 
@@ -481,7 +522,10 @@ text_date(PG_FUNCTION_ARGS)
 				dstr[MAXDATELEN + 1];
 
 	if (VARSIZE(str) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Bad date external representation (too long)");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_DATETIME_FORMAT),
+				 errmsg("invalid input syntax for type date: \"%s\"",
+						VARDATA(str))));
 
 	sp = VARDATA(str);
 	dp = dstr;
@@ -511,18 +555,22 @@ time_in(PG_FUNCTION_ARGS)
 	fsec_t		fsec;
 	struct tm	tt,
 			   *tm = &tt;
+	int			tz;
 	int			nf;
+	int			dterr;
 	char		lowstr[MAXDATELEN + 1];
 	char	   *field[MAXDATEFIELDS];
 	int			dtype;
 	int			ftype[MAXDATEFIELDS];
 
 	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad time external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-	 || (DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, NULL) != 0))
-		elog(ERROR, "Bad time external representation '%s'", str);
+		dterr = DTERR_BAD_FORMAT;
+	else
+		dterr = ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, &tz);
+	if (dterr != 0)
+		DateTimeParseError(dterr, str, "time");
 
 	tm2time(tm, fsec, &result);
 	AdjustTimeForTypmod(&result, typmod);
@@ -591,6 +639,43 @@ time_out(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		time_recv			- converts external binary format to time
+ *
+ * We make no attempt to provide compatibility between int and float
+ * time representations ...
+ */
+Datum
+time_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+
+#ifdef HAVE_INT64_TIMESTAMP
+	PG_RETURN_TIMEADT((TimeADT) pq_getmsgint64(buf));
+#else
+	PG_RETURN_TIMEADT((TimeADT) pq_getmsgfloat8(buf));
+#endif
+}
+
+/*
+ *		time_send			- converts time to binary format
+ */
+Datum
+time_send(PG_FUNCTION_ARGS)
+{
+	TimeADT		time = PG_GETARG_TIMEADT(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+#ifdef HAVE_INT64_TIMESTAMP
+	pq_sendint64(&buf, time);
+#else
+	pq_sendfloat8(&buf, time);
+#endif
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
+
 /* time_scale()
  * Adjust time type for specified scale factor.
  * Used by PostgreSQL type system to stuff columns.
@@ -642,17 +727,17 @@ AdjustTimeForTypmod(TimeADT *time, int32 typmod)
 #else
 	/* note MAX_TIME_PRECISION differs in this case */
 	static const double TimeScales[MAX_TIME_PRECISION + 1] = {
-		1,
-		10,
-		100,
-		1000,
-		10000,
-		100000,
-		1000000,
-		10000000,
-		100000000,
-		1000000000,
-		10000000000
+		1.0,
+		10.0,
+		100.0,
+		1000.0,
+		10000.0,
+		100000.0,
+		1000000.0,
+		10000000.0,
+		100000000.0,
+		1000000000.0,
+		10000000000.0
 	};
 #endif
 
@@ -661,9 +746,9 @@ AdjustTimeForTypmod(TimeADT *time, int32 typmod)
 		/*
 		 * Note: this round-to-nearest code is not completely consistent
 		 * about rounding values that are exactly halfway between integral
-		 * values.  On most platforms, rint() will implement round-to-nearest-even,
-		 * but the integer code always rounds up (away from zero).  Is it
-		 * worth trying to be consistent?
+		 * values.	On most platforms, rint() will implement
+		 * round-to-nearest-even, but the integer code always rounds up
+		 * (away from zero).  Is it worth trying to be consistent?
 		 */
 #ifdef HAVE_INT64_TIMESTAMP
 		if (*time >= INT64CONST(0))
@@ -673,8 +758,8 @@ AdjustTimeForTypmod(TimeADT *time, int32 typmod)
 		}
 		else
 		{
-			*time = - ((((- *time) + TimeOffsets[typmod]) / TimeScales[typmod])
-					   * TimeScales[typmod]);
+			*time = -((((-*time) + TimeOffsets[typmod]) / TimeScales[typmod])
+					  * TimeScales[typmod]);
 		}
 #else
 		*time = (rint(((double) *time) * TimeScales[typmod])
@@ -911,7 +996,9 @@ timestamp_time(PG_FUNCTION_ARGS)
 		PG_RETURN_NULL();
 
 	if (timestamp2tm(timestamp, NULL, tm, &fsec, NULL) != 0)
-		elog(ERROR, "Unable to convert timestamp to time");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
 #ifdef HAVE_INT64_TIMESTAMP
 
@@ -946,7 +1033,9 @@ timestamptz_time(PG_FUNCTION_ARGS)
 		PG_RETURN_NULL();
 
 	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
-		elog(ERROR, "Unable to convert timestamptz to time");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
 #ifdef HAVE_INT64_TIMESTAMP
 
@@ -1001,7 +1090,7 @@ time_interval(PG_FUNCTION_ARGS)
  * Convert interval to time data type.
  *
  * This is defined as producing the fractional-day portion of the interval.
- * Therefore, we can just ignore the months field.  It is not real clear
+ * Therefore, we can just ignore the months field.	It is not real clear
  * what to do with negative intervals, but we choose to subtract the floor,
  * so that, say, '-2 hours' becomes '22:00:00'.
  */
@@ -1022,7 +1111,7 @@ interval_time(PG_FUNCTION_ARGS)
 	}
 	else if (result < 0)
 	{
-		days = (-result + INT64CONST(86400000000-1)) / INT64CONST(86400000000);
+		days = (-result + INT64CONST(86400000000) - 1) / INT64CONST(86400000000);
 		result += days * INT64CONST(86400000000);
 	}
 #else
@@ -1161,7 +1250,10 @@ text_time(PG_FUNCTION_ARGS)
 				dstr[MAXDATELEN + 1];
 
 	if (VARSIZE(str) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Bad time external representation (too long)");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_DATETIME_FORMAT),
+				 errmsg("invalid input syntax for type time: \"%s\"",
+						VARDATA(str))));
 
 	sp = VARDATA(str);
 	dp = dstr;
@@ -1192,9 +1284,12 @@ time_part(PG_FUNCTION_ARGS)
 				lowunits[MAXDATELEN + 1];
 
 	if (VARSIZE(units) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "TIME units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("\"time\" units \"%s\" not recognized",
+						DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 	up = VARDATA(units);
 	lp = lowunits;
 	for (i = 0; i < (VARSIZE(units) - VARHDRSZ); i++)
@@ -1259,9 +1354,12 @@ time_part(PG_FUNCTION_ARGS)
 			case DTK_CENTURY:
 			case DTK_MILLENNIUM:
 			default:
-				elog(ERROR, "TIME units '%s' not supported",
-					 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("\"time\" units \"%s\" not recognized",
+							 DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 				result = 0;
 		}
 	}
@@ -1275,9 +1373,11 @@ time_part(PG_FUNCTION_ARGS)
 	}
 	else
 	{
-		elog(ERROR, "TIME units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("\"time\" units \"%s\" not recognized",
+						DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
 		result = 0;
 	}
 
@@ -1321,18 +1421,20 @@ timetz_in(PG_FUNCTION_ARGS)
 			   *tm = &tt;
 	int			tz;
 	int			nf;
+	int			dterr;
 	char		lowstr[MAXDATELEN + 1];
 	char	   *field[MAXDATEFIELDS];
 	int			dtype;
 	int			ftype[MAXDATEFIELDS];
 
 	if (strlen(str) >= sizeof(lowstr))
-		elog(ERROR, "Bad time with time zone"
-			 " external representation (too long) '%s'", str);
-
-	if ((ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf) != 0)
-	  || (DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, &tz) != 0))
-		elog(ERROR, "Bad time external representation '%s'", str);
+		dterr = DTERR_BAD_FORMAT;
+	else
+		dterr = ParseDateTime(str, lowstr, field, ftype, MAXDATEFIELDS, &nf);
+	if (dterr == 0)
+		dterr = DecodeTimeOnly(field, ftype, nf, &dtype, tm, &fsec, &tz);
+	if (dterr != 0)
+		DateTimeParseError(dterr, str, "time with time zone");
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 	tm2timetz(tm, fsec, tz, result);
@@ -1359,6 +1461,47 @@ timetz_out(PG_FUNCTION_ARGS)
 	PG_RETURN_CSTRING(result);
 }
 
+/*
+ *		timetz_recv			- converts external binary format to timetz
+ */
+Datum
+timetz_recv(PG_FUNCTION_ARGS)
+{
+	StringInfo	buf = (StringInfo) PG_GETARG_POINTER(0);
+	TimeTzADT  *time;
+
+	time = (TimeTzADT *) palloc(sizeof(TimeTzADT));
+
+#ifdef HAVE_INT64_TIMESTAMP
+	time->time = pq_getmsgint64(buf);
+#else
+	time->time = pq_getmsgfloat8(buf);
+#endif
+	time->zone = pq_getmsgint(buf, sizeof(time->zone));
+
+	PG_RETURN_TIMETZADT_P(time);
+}
+
+/*
+ *		timetz_send			- converts timetz to binary format
+ */
+Datum
+timetz_send(PG_FUNCTION_ARGS)
+{
+	TimeTzADT  *time = PG_GETARG_TIMETZADT_P(0);
+	StringInfoData buf;
+
+	pq_begintypsend(&buf);
+#ifdef HAVE_INT64_TIMESTAMP
+	pq_sendint64(&buf, time->time);
+#else
+	pq_sendfloat8(&buf, time->time);
+#endif
+	pq_sendint(&buf, time->zone, sizeof(time->zone));
+	PG_RETURN_BYTEA_P(pq_endtypsend(&buf));
+}
+
+
 /* timetz2tm()
  * Convert TIME WITH TIME ZONE data type to POSIX time structure.
  */
@@ -1414,20 +1557,12 @@ timetz_scale(PG_FUNCTION_ARGS)
 static int
 timetz_cmp_internal(TimeTzADT *time1, TimeTzADT *time2)
 {
-	/* Primary sort is by true (GMT-equivalent) time */
-#ifdef HAVE_INT64_TIMESTAMP
-	int64		t1,
-				t2;
-
-	t1 = time1->time + (time1->zone * INT64CONST(1000000));
-	t2 = time2->time + (time2->zone * INT64CONST(1000000));
-#else
 	double		t1,
 				t2;
 
+	/* Primary sort is by true (GMT-equivalent) time */
 	t1 = time1->time + time1->zone;
 	t2 = time2->time + time2->zone;
-#endif
 
 	if (t1 > t2)
 		return 1;
@@ -1530,12 +1665,13 @@ timetz_larger(PG_FUNCTION_ARGS)
 {
 	TimeTzADT  *time1 = PG_GETARG_TIMETZADT_P(0);
 	TimeTzADT  *time2 = PG_GETARG_TIMETZADT_P(1);
+	TimeTzADT  *result;
 
-	if (DatumGetBool(DirectFunctionCall2(timetz_gt,
-										 TimeTzADTPGetDatum(time1),
-										 TimeTzADTPGetDatum(time2))))
-		PG_RETURN_TIMETZADT_P(time1);
-	PG_RETURN_TIMETZADT_P(time2);
+	if (timetz_cmp_internal(time1, time2) > 0)
+		result = time1;
+	else
+		result = time2;
+	PG_RETURN_TIMETZADT_P(result);
 }
 
 Datum
@@ -1543,12 +1679,13 @@ timetz_smaller(PG_FUNCTION_ARGS)
 {
 	TimeTzADT  *time1 = PG_GETARG_TIMETZADT_P(0);
 	TimeTzADT  *time2 = PG_GETARG_TIMETZADT_P(1);
+	TimeTzADT  *result;
 
-	if (DatumGetBool(DirectFunctionCall2(timetz_lt,
-										 TimeTzADTPGetDatum(time1),
-										 TimeTzADTPGetDatum(time2))))
-		PG_RETURN_TIMETZADT_P(time1);
-	PG_RETURN_TIMETZADT_P(time2);
+	if (timetz_cmp_internal(time1, time2) < 0)
+		result = time1;
+	else
+		result = time2;
+	PG_RETURN_TIMETZADT_P(result);
 }
 
 /* timetz_pl_interval()
@@ -1798,7 +1935,9 @@ timestamptz_timetz(PG_FUNCTION_ARGS)
 		PG_RETURN_NULL();
 
 	if (timestamp2tm(timestamp, &tz, tm, &fsec, &tzn) != 0)
-		elog(ERROR, "Unable to convert timestamptz to timetz");
+		ereport(ERROR,
+				(errcode(ERRCODE_DATETIME_VALUE_OUT_OF_RANGE),
+				 errmsg("timestamp out of range")));
 
 	result = (TimeTzADT *) palloc(sizeof(TimeTzADT));
 
@@ -1874,7 +2013,10 @@ text_timetz(PG_FUNCTION_ARGS)
 				dstr[MAXDATELEN + 1];
 
 	if (VARSIZE(str) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Bad timetz external representation (too long)");
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_DATETIME_FORMAT),
+		   errmsg("invalid input syntax for type time with time zone: \"%s\"",
+				  VARDATA(str))));
 
 	sp = VARDATA(str);
 	dp = dstr;
@@ -1905,9 +2047,12 @@ timetz_part(PG_FUNCTION_ARGS)
 				lowunits[MAXDATELEN + 1];
 
 	if (VARSIZE(units) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "TIMETZ units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("\"time with time zone\" units \"%s\" not recognized",
+						DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 	up = VARDATA(units);
 	lp = lowunits;
 	for (i = 0; i < (VARSIZE(units) - VARHDRSZ); i++)
@@ -1931,17 +2076,18 @@ timetz_part(PG_FUNCTION_ARGS)
 		switch (val)
 		{
 			case DTK_TZ:
-				result = tz;
+				result = -tz;
 				break;
 
 			case DTK_TZ_MINUTE:
-				result = tz / 60;
-				TMODULO(result, dummy, 60e0);
+				result = -tz;
+				result /= 60;
+				FMODULO(result, dummy, 60e0);
 				break;
 
 			case DTK_TZ_HOUR:
-				dummy = tz;
-				TMODULO(dummy, result, 3600e0);
+				dummy = -tz;
+				FMODULO(dummy, result, 3600e0);
 				break;
 
 			case DTK_MICROSEC:
@@ -1985,25 +2131,31 @@ timetz_part(PG_FUNCTION_ARGS)
 			case DTK_CENTURY:
 			case DTK_MILLENNIUM:
 			default:
-				elog(ERROR, "TIMETZ units '%s' not supported",
-					 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+				ereport(ERROR,
+						(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+						 errmsg("\"time with time zone\" units \"%s\" not recognized",
+							 DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 				result = 0;
 		}
 	}
 	else if ((type == RESERV) && (val == DTK_EPOCH))
 	{
 #ifdef HAVE_INT64_TIMESTAMP
-		result = ((time->time / 1000000e0) + time->zone);
+		result = ((time->time / 1000000e0) - time->zone);
 #else
-		result = (time->time + time->zone);
+		result = (time->time - time->zone);
 #endif
 	}
 	else
 	{
-		elog(ERROR, "TIMETZ units '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-											   PointerGetDatum(units))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("\"time with time zone\" units \"%s\" not recognized",
+						DatumGetCString(DirectFunctionCall1(textout,
+											 PointerGetDatum(units))))));
+
 		result = 0;
 	}
 
@@ -2028,9 +2180,12 @@ timetz_zone(PG_FUNCTION_ARGS)
 				lowzone[MAXDATELEN + 1];
 
 	if (VARSIZE(zone) - VARHDRSZ > MAXDATELEN)
-		elog(ERROR, "Time zone '%s' not recognized",
-			 DatumGetCString(DirectFunctionCall1(textout,
-												 PointerGetDatum(zone))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("time zone \"%s\" not recognized",
+						DatumGetCString(DirectFunctionCall1(textout,
+											  PointerGetDatum(zone))))));
+
 	up = VARDATA(zone);
 	lp = lowzone;
 	for (i = 0; i < (VARSIZE(zone) - VARHDRSZ); i++)
@@ -2062,7 +2217,10 @@ timetz_zone(PG_FUNCTION_ARGS)
 	}
 	else
 	{
-		elog(ERROR, "Time zone '%s' not recognized", lowzone);
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("time zone \"%s\" not recognized", lowzone)));
+
 		PG_RETURN_NULL();
 	}
 
@@ -2081,9 +2239,11 @@ timetz_izone(PG_FUNCTION_ARGS)
 	int			tz;
 
 	if (zone->month != 0)
-		elog(ERROR, "INTERVAL time zone '%s' not legal (month specified)",
-			 DatumGetCString(DirectFunctionCall1(interval_out,
-												 PointerGetDatum(zone))));
+		ereport(ERROR,
+				(errcode(ERRCODE_INVALID_PARAMETER_VALUE),
+				 errmsg("\"interval\" time zone \"%s\" not valid",
+						DatumGetCString(DirectFunctionCall1(interval_out,
+											  PointerGetDatum(zone))))));
 
 #ifdef HAVE_INT64_TIMESTAMP
 	tz = -(zone->time / INT64CONST(1000000));
